#include <stdlib.h>
#include <iostream.h>
#include "TwoThree.h"

// ---------------------------------------------------------------- Definitions

   // Return the minumum of two numbers
inline static int
MIN(int a, int b) {
   return  (a < b) ? a : b;
}

   // Return the maximum of two numbers
inline static int
MAX(int a, int b) {
   return  (a > b) ? a : b;
}

   // Return the minumum of three numbers
inline static int
MIN(int a, int b, int c) {
   return  (a < b) ? MIN(a, c) : MIN(a, b);
}

   // Return the maximum of two numbers
inline static int
MAX(int a, int b, int c) {
   return  (a > b) ? MAX(a, c) : MAX(a, b);
}

   // Exchange two items
inline static void
ExchangeItems(ItemType * & item1, ItemType * & item2) {
   ItemType * item = item1;
   item1 = item2;
   item2 = item;
}

   // Exchange two trees
inline static void
ExchangeTrees(TwoThreeNode * & tree1, TwoThreeNode * & tree2) {
   TwoThreeNode * tree = tree1;
   tree1 = tree2;
   tree2 = tree;
}

// ----------------------------------------------- Constructors and Destructors

TwoThreeNode::TwoThreeNode(ItemType * item1, ItemType * item2)
   : myNumKeys(0)
{
   myData[0] = item1;
   myData[1] = item2;
   if (myData[0])  ++myNumKeys;
   if (myData[1])  ++myNumKeys;
   Reset();
}

TwoThreeNode::~TwoThreeNode(void) {
   for (int i = 0; i < MAX_SUBTREES; ++i) {
      if (mySubtree[i])  delete  mySubtree[i] ;
   }
}

// ----------------------------------------------------- Restructuring

void
TwoThreeNode::SplitNode(ItemType * & item, int  index,
                                           TwoThreeNode * & overflow)
{
   // Split this node -- put the median data item into "item" and
   // let "overflow" point at the new node.

      // Sift the median data-element into "item"
   if (index < MIDDLE) {
      ExchangeItems(item, myData[LEFT]);
         // exchange overflow with middle subtree
      ExchangeTrees(overflow, mySubtree[MIDDLE]);
   } else if (index > MIDDLE) {
      ExchangeItems(item, myData[MIDDLE]);
         // exchange overflow with right subtree
      ExchangeTrees(overflow, mySubtree[RIGHT]);
   }

      // Remove rightmost key and its subtree from this node and
      // put them into a newly created node.
   TwoThreeNode * newNode = new TwoThreeNode(myData[--myNumKeys]);
   myData[myNumKeys] = NULL;
   newNode->mySubtree[MIDDLE] = mySubtree[MAX_KEYS];
         // If overflow and the middle subtree are both non-NUll, then we
         // need to figure out which of the two goes with the new node and
         // which one stays with the old node.
   if (overflow  &&  mySubtree[MIDDLE]  && 
       (mySubtree[MIDDLE]->Compare(overflow->myData[LEFT]->Key(), index)
                                                                 == MIN_CMP)) {
         // Middle subtree goes with the new node
      newNode->mySubtree[LEFT] = mySubtree[MIDDLE];
      mySubtree[MIDDLE] = overflow;
   } else {
      newNode->mySubtree[LEFT]   = overflow;
   }
   mySubtree[MAX_KEYS] = NULL;
      // Set overflow to the new node
   overflow = newNode;
}

void
TwoThreeNode::InternalInsert(ItemType * & item, int  index,
                                            TwoThreeNode * & overflow)
{
   int i = 0;
   if (myNumKeys < MAX_KEYS) {
         // We have a 2-node with room for another key
      if (index < (MAX_KEYS - 1)) {
            // Shift keys and subtrees over to make room
         for (i = myNumKeys; i > index; --i) {
            myData[i] = myData[i - 1];
            mySubtree[i + 1] = mySubtree[i];
         }
      }
      myData[index] = item;
      mySubtree[index + 1] = overflow;
      ++myNumKeys;
      overflow = NULL;
   } else {
         // We have an overflow, need to split this node
      SplitNode(item, index, overflow);
   }
}

void
TwoThreeNode::InternalDelete(int  index)
{
   int  nonLeaf = (mySubtree[LEFT] != NULL);   // See if this is a leaf

      // Shift all the data items and their subtrees to the left
      // (shift subtrees too if there are any)
   int i;
   for (i = index; i < myNumKeys - 1; ++i) {
      myData[i] = myData[i + 1];
      if (nonLeaf)  mySubtree[i] = mySubtree[i + 1];
   }
   if (nonLeaf) {
         // Shift rightmost subtree to the left
      if (index == 0)  mySubtree[myNumKeys - 1] = mySubtree[myNumKeys];
      mySubtree[myNumKeys] = NULL;
   }
   myData[--myNumKeys] = NULL;  // remove "old" rightmost key
}

void
TwoThreeNode::MergeNodes(int  index, TwoThreeNode * underflow,
                                     TwoThreeNode * sibling)
{
   // NOTE: when merging a 2-node with its deficient 2-node sibling,
   //       we shall choose to *always* merge values into the leftmost
   //       of the two subtrees. This way, if a node is deficient, we
   //       know that it is either a leaf, or that the deficiency was
   //       propagated up, in which case its ONLY non-empty subtree
   //       its leftmost subtree.

   int i;
   if (index == 0) {
      // Sibling is to the right of underflow, merge into underflow since
      // it is the leftmost of the two.

         // First absorb key from parent
      underflow->myData[LEFT] = myData[index];
         // Now absorb key from sibling
      underflow->myData[MIDDLE] = sibling->myData[LEFT];
      underflow->myNumKeys = MAX_KEYS;
         // Now absorb left & middle subtrees of sibling
      for (i = LEFT; i <= MIDDLE; ++i) {
         underflow->mySubtree[i + 1] = sibling->mySubtree[i];
      }
         // Now discard sibling and remove the parent's link to it
      for (i = 0; i < MAX_SUBTREES; i++)  sibling->mySubtree[i] = NULL;
      delete  sibling;
   } else {
      // Sibling is to the left of underflow, merge into sibling since
      // it is the leftmost of the two.

         // First absorb key from parent
      sibling->myData[MIDDLE] = myData[--index];
      sibling->myNumKeys = MAX_KEYS;
         // Now absorb left subtree of underflow
      sibling->mySubtree[RIGHT] = underflow->mySubtree[LEFT];
         // Now discard underflow and remove the parent's link to it
      for (i = 0; i < MAX_SUBTREES; i++)  underflow->mySubtree[i] = NULL;
      delete  underflow;
   }
      // Remove copied key from parent
   if (--myNumKeys > 0) {
      if (index < myNumKeys) {
         // Shift keys and subtrees from right to left
         for (i = index; i < myNumKeys; ++i) {
            myData[i] = myData[i + 1];
            mySubtree[i + 1] = mySubtree[i + 2];
         }
      }
   } 
   myData[myNumKeys] = NULL;
   mySubtree[myNumKeys + 1] = NULL;
}

void
TwoThreeNode::Redistribute(int  index)
{
      // First figure out sibling to try and borrow from
   TwoThreeNode * underflow = mySubtree[index];
   TwoThreeNode * sibling   = mySubtree[(index == 0) ? index+1 : index-1];
   if (sibling->myNumKeys <= MEDIAN_INDEX) {
         // The sibling has no data-item to spare -- merge with it
      MergeNodes(index, underflow, sibling);
   } else {
         // The sibling has a data-item to spare -- borrow from it
      if (index == 0) {
         // Sibling is to the right
            // Borrow item from parent
         underflow->myData[underflow->myNumKeys++] = myData[index];
            // Parent gets min item from right subtree (sibling)
         myData[index] = sibling->myData[LEFT];
            // Absorb left subtree of sibling into right subtree of underflow
         underflow->mySubtree[underflow->myNumKeys] = sibling->mySubtree[LEFT];
            // Shift sibling keys and subtrees to the left
         sibling->InternalDelete(LEFT);
      } else {
         // Sibling is to the left
            // Borrow item from parent
         underflow->myData[underflow->myNumKeys++] = myData[--index];
            // Parent gets max item from left subtree (sibling)
         myData[index] = sibling->myData[sibling->myNumKeys-1];
            // Middle subtree points to old left subtree
         underflow->mySubtree[MIDDLE] = underflow->mySubtree[LEFT];
            // Absorb right subtree of sibling into left subtree of underflow
         underflow->mySubtree[LEFT] = sibling->mySubtree[sibling->myNumKeys];
            // Discard rightmost tree and subtree of sibling
         sibling->InternalDelete(sibling->myNumKeys-1);
      }
   }
}

// ------------------------------------------------------ Search/Insert/Delete

cmp_t
TwoThreeNode::Compare(KeyType key, int & index, cmp_t cmp) const
{
   cmp_t  result;

   switch (cmp) {
      case EQ_CMP :  // Standard comparison
         index = 0;
         result = myData[index]->Compare(key);
         while ((result == MAX_CMP)  &&  (++index < myNumKeys)) {
            result = myData[index]->Compare(key);
         }
         return  result;

      case MIN_CMP :  // Find the minimal element in this tree
         index = 0;
         return  (mySubtree[index] == NULL) ? EQ_CMP : MIN_CMP;

      case MAX_CMP :  // Find the maximal element in this tree
         index = myNumKeys;
         if (mySubtree[index] == NULL) {
            --index;  // get index of max-key
            return  EQ_CMP;
         } else {
            return  MAX_CMP;
         }
   }
}

ItemType *
TwoThreeNode::Search23(KeyType key, TwoThreeNode * root, cmp_t cmp)
{
   int    index  = 0;
   cmp_t  result;
   while (root  &&  (result = root->Compare(key, index, cmp))) {
      root = root->mySubtree[index];
   }
   return  (root) ? root->myData[index] : NULL;
}

ItemType *
TwoThreeNode::Insert23(ItemType * item, TwoThreeNode * & root) {
   TwoThreeNode * overflow = NULL;
   ItemType     * found = Insert23(item, root, overflow);
      // See if we need to "grow" a new root due to overflow
   if (overflow) {
      TwoThreeNode * newRoot = new TwoThreeNode(item);
      newRoot->mySubtree[MIDDLE] = overflow;
      newRoot->mySubtree[LEFT]   = root;
      root = newRoot;
   }
   return  found;
}

ItemType *
TwoThreeNode::Delete23(KeyType key, TwoThreeNode * & root, cmp_t cmp) {
   TwoThreeNode * underflow = NULL;
   ItemType * found = Delete23(key, root, underflow, cmp);
   if (root  &&  underflow) {
         // Discard the root and use it's left subtree instead
      TwoThreeNode * oldRoot = root;
      root = root->mySubtree[LEFT];
      delete  oldRoot;
   }
   return  found;
}

ItemType *
TwoThreeNode::Insert23(ItemType * & item, TwoThreeNode * & root,
                                          TwoThreeNode * & overflow)
{
   overflow = NULL;

      // If root is NULL, the tree is empty so make the first node & quit
   if (root == NULL) {
      root = new TwoThreeNode(item);
      return  NULL;
   }

      // Perform comparison to see where key fits
   int    index = 0;
   cmp_t  result = root->Compare(item->Key(), index);
   if (result == EQ_CMP)  return  root->myData[index];  // Already in tree!

      // See if we need to insert into a subtree
   if (root->mySubtree[index]) {
         // Yes, insert into index-th subtree
      ItemType * found = Insert23(item, root->mySubtree[index], overflow);
         // Quit now if the key was found, or it was inserted without overflow
      if (found || !overflow)  return  found;
         // We have an overflow - recompute index for the new key
      result = root->Compare(item->Key(), index);
   }
      // Insert key into this node
   root->InternalInsert(item, index, overflow);
   return  NULL;
}


ItemType *
TwoThreeNode::Delete23(KeyType key, TwoThreeNode * & root,
                                    TwoThreeNode * & underflow, cmp_t cmp)
{
   ItemType * found = NULL;
   underflow = NULL;

      // If root is NULL then key isnt in the tree
   if (root == NULL)  return  NULL;

      // Perform comparison to see where key fits
   int    index = 0;
   cmp_t  result = root->Compare(key, index, cmp);
   if (result != EQ_CMP) {
         // Delete from a subtree
      found = Delete23(key, root->mySubtree[index], underflow, cmp);
   } else {
         // The key is right here in the index-th data field
      found = root->myData[index];
      if (root->mySubtree[LEFT]) {
            // This node is an interior (non-leaf) node. Replace the item
            // to delete with its predecessor in the tree
         root->myData[index] = Delete23(key, root->mySubtree[index],
                                        underflow, MAX_CMP);
      } else {
            // This node is a leaf - so just remove the index-th item
         root->InternalDelete(index);
      }
   }

      // Check for underflow during deletion from index-th subtree,
      // if it occurred, need to redistribute data
   if (underflow)  root->Redistribute(index);
      // Propagate underflow to parent if this node is now deficient
   underflow = (root->myNumKeys == 0) ? root : NULL;
   return  found;
}

// --------------------------------------------------------------- Verification

int
TwoThreeNode::Height() const {
   int  leftHeight   = (mySubtree[LEFT])   ? mySubtree[LEFT]->Height()   : 0;
   int  middleHeight = (mySubtree[MIDDLE]) ? mySubtree[MIDDLE]->Height() : 0;
   int  rightHeight  = (mySubtree[RIGHT])  ? mySubtree[RIGHT]->Height()  : 0;
   return  (1 + MAX(leftHeight, middleHeight, rightHeight));
}

int
TwoThreeNode::Check23() const {
   int  valid = 1;

      // First verify that subtrees are correct
   if (mySubtree[LEFT])   valid *= mySubtree[LEFT]->Check23();
   if (mySubtree[MIDDLE]) valid *= mySubtree[MIDDLE]->Check23();
   if (mySubtree[RIGHT])  valid *= mySubtree[RIGHT]->Check23();

      // First get the height of each subtree
   int  leftHeight   = (mySubtree[LEFT])   ? mySubtree[LEFT]->Height()   : 0;
   int  middleHeight = (mySubtree[MIDDLE]) ? mySubtree[MIDDLE]->Height() : 0;
   int  rightHeight  = (mySubtree[RIGHT])  ? mySubtree[RIGHT]->Height()  : 0;

      // Verify that 2-3 tree property is satisfied
   if ((leftHeight != middleHeight) ||
       ((myNumKeys == MAX_KEYS) && (middleHeight != rightHeight))) {
      valid = 0;
      cerr << "Different subtree heights (" 
           << leftHeight << " " << middleHeight << " " << rightHeight
           << ") at " << myNumKeys+1 << "-node " << Key(LEFT) << endl;
   }

   int i = 0;

      // Check that keyfields are in proper sequence
   for (i = myNumKeys - 1; i >= 0; --i) {
         // Check order of key[i] and its predecessor key
      if ((i > 0) && (myData[i-1]->Compare(myData[i]->Key()) == MIN_CMP))
      {
         valid = 0;
         cerr << "key[" << i << "]=" << *myData[i] << " is *smaller* than "
              << "key[" << i-1 << "]=" << *myData[i-1]
              << " at " << myNumKeys+1 << "-node " << Key(LEFT) << endl;
      }
         // Check that key[i] is <= its right subtree's smallest key
      if (mySubtree[i+1] &&
          (myData[i]->Compare(mySubtree[i+1]->Key(LEFT)) == MIN_CMP))
      {
         valid = 0;
         cerr << "key[" << i << "]=" << *myData[i] << " is *greater* than "
              << "its subtree[" << i+1 << "]'s smallest key"
              << " at " << myNumKeys+1 << "-node " << Key(LEFT) << endl;
      }
         // Check that key[i] is >= its left subtree's largest key
      dir_t  maxKey = dir_t((mySubtree[i]) ? mySubtree[i]->NumKeys() - 1 : 0);
      if (mySubtree[i] &&
          (myData[i]->Compare(mySubtree[i]->Key(maxKey)) == MAX_CMP))
      {
         valid = 0;
         cerr << "key[" << i << "]=" << *myData[i] << " is *smaller* than "
              << "its subtree[" << i << "]'s greatest key"
              << " at " << myNumKeys+1 << "-node " << Key(LEFT) << endl;
      }
   }

   return  valid;
}

//----------------------------------------------- Routines for dumping the tree

static ostream &
Indent(ostream & os, int len) {
   for (int i = 0; i < len; i++) {
      os << ' ';
   }
   return  os;
}

enum TraversalOrder { LTREE, LKEY, MTREE, RKEY, RTREE };

static void
Dump23(ostream & os, TraversalOrder order,
                           const TwoThreeNode * node, unsigned  level=0)
{
    unsigned  len = (level * 5) + 1;
    if ((order == LTREE) && (node->Subtree(TwoThreeNode::LEFT) == NULL)) {
       //Indent(os, len) << "     **NULL**" << endl;
    }
    if (order == LKEY) {
       Indent(os, len) << node->Key(TwoThreeNode::LEFT) << endl;
    }
    if ((order == MTREE) && (node->Subtree(TwoThreeNode::MIDDLE) == NULL)) {
       //Indent(os, len) << "     **NULL**" << endl;
    }
    if (node->NumKeys() < 2)  return;
    if (order == RKEY) {
       Indent(os, len) << node->Key(TwoThreeNode::RIGHT) << endl;
    }
    if ((order == RTREE) && (node->Subtree(TwoThreeNode::RIGHT) == NULL)) {
       //Indent(os, len) << "     **NULL**" << endl;
    }
}

static void
Dump23(ostream & os, const TwoThreeNode * node, unsigned  level=0)
{
   if (node == NULL) {
      os << "***EMPTY TREE***" << endl;
   } else {
      if (node->NumKeys() > 1) {
         Dump23(os, RTREE, node, level);
         if (node->Subtree(TwoThreeNode::RIGHT)  !=  NULL) {
            Dump23(os, node->Subtree(TwoThreeNode::RIGHT), level+1);
         }
         Dump23(os, RKEY, node, level);
      }
      if (node->Subtree(TwoThreeNode::MIDDLE)  !=  NULL) {
         Dump23(os, node->Subtree(TwoThreeNode::MIDDLE), level+1);
      }
      Dump23(os, MTREE, node, level);
      Dump23(os, LKEY, node, level);
      if (node->Subtree(TwoThreeNode::LEFT)  !=  NULL) {
         Dump23(os, node->Subtree(TwoThreeNode::LEFT), level+1);
      }
      Dump23(os, LTREE, node, level);
   }// if non-empty tree
}

void
TwoThreeTree::DumpTree(ostream & os) const {
   Dump23(os, myRoot);
}

const char TwoThreeTree::ourKind[] = "2-3 tree";
